import cv2
import random
import sys
import os
import math
import cozmo
from numba import jit
import numpy as np
import pandas as pd
import MotionUpdate
import Variables as varia
	
#See Stitching for why this is here.
#Reenable this @jit call if you ever rework this method to use samples much wider than one column of pixels; i.e., on the scale of full images. The pause that the computer takes to send all of this stuff to the GPU is long enough for the CPU to still finish before the GPU would be done messing with it - at least, that's how it is on my machine.
#Essentially, the boost in power from using a GPU would only be helpful if you're dealing with large imgages.
#@jit
def localize():
	"""Performs calculations in order to figure out where it is facing in relation to the panorama image it's already taken."""
	
	print('Begin localize')

	#Retrieves the paths from Variables.py
	picturePath = varia.getPicturePath()
	dataPath = varia.getDataPath()
	homePic = cv2.imread(picturePath + 'latestImage.jpeg')
	pano = cv2.imread(picturePath + 'Panorama.jpeg')
 
	#Store the dimensions of the home picture and panorama for (assumedly) faster accessing
	homeDims = homePic.shape
	panoDims = pano.shape
	
	#Sets the smallest height for images to be taken from the panorama to be

	heightDiff = homePic.shape[1] - pano.shape[1]

	#Crop the home picture just in case the panorama is shorter in some way
	homePic = homePic[int(heightDiff / 2):homePic.shape[0] - int(heightDiff / 2), 0:1]
	
	#Crop either the homePic or panorama to the minimum height between the two if one is taller than the other because I didn't have time to write code to check for an integer rounding error that results in the heights of the home picture and panorama not being the same or whatever the proper name for that would be
	minHeight = min(homeDims[0], panoDims[0])
	if(pano.shape[0] > minHeight):
		pano = pano[0:minHeight, 0:panoDims[1]]
	if(homePic.shape[0] > minHeight):
		homePic = homePic[0:minHeight, 0:homeDims[1]]
	
	#Set these again; may not have to do so
	homeDims = homePic.shape
	panoDims = pano.shape
	
	print('Shapes set')
	#When resampling after the first random turn, all of the points remaining from the initial random 1000 batch should "turn" the same number of degrees, with some noise.
	
	#Generate the array for starting pixel values for which to take MSE samples of
	pointChecks = []
	maxSamples = 1000
	used = np.zeros(panoDims[1], dtype=bool)
	
	#print('panoWidth: ' + str(panoDims[1]))
	#print (used)
	#This check is mostly here to see if it'd be worth it to take a clean 1000 samples or just grab a sample from every column of pixels of the panorama; you shouldn't get to the 1000 branch them unless you don't downres your panorama enough. The 1000 value used is arbitrary, also; I just figured that any panorama created from the images taken by a Cozmo robot's camera wouldn't be much larger than 1000 pixels wide.
	if maxSamples > (panoDims[1]):
		for i in range(panoDims[1]):

			pointChecks.append(i)
		#Shuffle these afterwards to get randomly placed choices; not sure if doing this actually matters.
		random.shuffle(pointChecks)
		#Create dataframe object to store data

	else:
		for i in range(maxSamples):
			#Reroll the sampled pixel if we've seen it before. I figured it'd be best to do that just to get a good sample population.

			current = random.randint(0, panoDims[1] - 1)
			while used[current] == True:
				#print('Rerolling')
				current = random.randint(0, panoDims[1] - 1)

			#print('Sampling from pixel ' + str(current))
			pointChecks.append(current)
			used[current] = True
		#Create dataframe object to store data
		df = pd.DataFrame(pointChecks, columns=['Population Pixel'])
		df['Probabilities'] = np.ones(maxSamples)
		
	print('Points selected')
	
	#Create dataframe object to store data
	df = pd.DataFrame(pointChecks, columns=['Population Pixel'])
	print('pointChecks length: ' + str(len(pointChecks)))
	df['Probabilities'] = np.ones(len(pointChecks))
	print('Points selected')
	
	iters = 1
	final = -1
	while final < 0:
		print('Running localize iter ' + str(iters))
		input('Press enter to start localization iter ' + str(iters))
		#Grab the first image taken by the robot after randomly turning and calculate the mse values
		cozmo.run_program(rotato)
		update = int(((varia.getUpdate()) / 360) * panoDims[1])
		
		homePic = cv2.imread(picturePath + 'latestImage.jpeg')
		homeDims = homePic.shape
		
		#Cropjob again
		homePic = homePic[int(heightDiff / 2):homeDims[0] - int(heightDiff / 2), 0:1]

		minHeight = min(homeDims[0], panoDims[0])
		if(pano.shape[0] > minHeight):
			pano = pano[0:minHeight, 0:panoDims[1]]
		if(homePic.shape[0] > minHeight):
			homePic = homePic[0:minHeight, 0:homeDims[1]]
		
		#Update poses via adding the approximate number of pixel advanced via rotation, while also accounting for wraparound
		#Append two images as numpy arrays to not have to use columns
		for i in range(len(df['Population Pixel'])):
			df.at[i, 'Population Pixel'] = ((df.loc[i, 'Population Pixel'] + update) % panoDims[1])

		#print('homePic height: ' + str(homePic.shape[0]))
		#print('minHeight: ' + str(minHeight))
		
		#Get the sample pictures at those values (columns instead of pictures now)
		homeFloatRep = homePic.astype("float")
		MSEList = []
		for randomPoint in pointChecks:
			#The 'Mean Squared Error' between the two images is the sum of the squared difference between the two images.
			#NOTE: the two images must have the same dimensions.
			MSEList.append((np.sum((homeFloatRep - pano[0:minHeight, randomPoint:randomPoint + 1].astype("float")) ** 2)) / float(homePic.shape[0] * pano.shape[1]))
			
		
		df['MSE'] = MSEList
		#print('MSE values calculated and stored')
		#print(df['MSE'])


		#Get the sum of the picture's light values (MSE)
		MSESum = sum(df['MSE'])

		#Normalize and invert values
		df['Normalized MSE'] = (df['MSE']/MSESum)
		df['Inverted'] = (1/df['Normalized MSE'])
		#print(df['Probabilities'])
		#Normalize the values
		newSum = sum(df['Inverted'])
		#print('\n\n\n\n\n\nBefore: ')
		#print(df['Probabilities'])
		df['Probabilities'] = (df['Inverted'] / newSum)
		
		#Old code for checking whether or not the probabilities were being computed correctly. Uncomment if you wish for the console to spit out a bunch of information at you.
		#for i in range(len(df['Probabilities'])):
			##df.at[i, 'Probabilities'] *= (df.loc[i, 'Inverted'] /  newSum)
			#df.at[i, 'Probabilities'] = (df.loc[i, 'Inverted'] /  newSum)
		#print('\n\n\n\n\n\nAfter: ')
		#print(df['Probabilities'])
		
		#This block of code is a way to calculate the cumulative probabilities of the current population members remaining in the next iteration.
		#Here is a small Ascii example of what happens:
		#
		#Population: 				[1]    [7]    [2]    [3]    [5]	These are the relative likelihoods of each member representing the point the robot is looking at.
		#
		#Individual Probabilities:		[.056] [.389] [.111] [.167] [.278]	These are the probabilities of each member being selected to remain once normalized
		#										(put over 18, in this case).
		#Cumulative Probabilities:		[.056] [.445] [.556] [.723] [1]	These are the cumulative probabilities used to select new members; it kind of works like the
		#										Fibonacci sequence, except you add up all prior members, add the value of the current cell, and put that
		#										into the cell of the cumulative probabilites array with the same index.
		#
		#										Do take note that if you were to add all of the cumulative probabilites I've listed together, you'd get
		#										1.001 at the end. You'll get similar imprecision with floating point addition. To fix this, just set the
		#										last value should be just 1 flat.
		hold = 0
		cumProbs = []
		length = len(df['Probabilities'])
		#print('panoWidth: ' + str(panoDims[1]))
		for i in range(int(length)):
			cumProbs.append((df.loc[i, 'Probabilities']) + hold)
			hold = cumProbs[i]
		#This set to 1 is done to avoid any errors that might come about as a result of floating point errors in the above calculation - just a failsafe. If you'd like to check what is actually saved before you do this, reenable this print statement.
		#print(cumProbs)
		cumProbs[length - 1] = 1
		#print('Probs length: ' + str(length))
		df['Cumulative Probs'] = cumProbs
		#df.iat[maxSamples - 1, 'Cumulative Probs'] = 1
		#This'll display the cumulative probabilities after calculation. Reenable if desired.
		#print(cumProbs)

		latest_image = cv2.imread(picturePath + 'latestImage.jpeg')
		
		#Take a guess x, generate a random integer between 0 and the number of samples remaining; if the value of Cumulative Probs[i] < x's probability, keep x; otherwise, discard x
		#Rinse and repeat for all samples remaining to be compared with the latest image
		#If there are more than one guesses remaining, take a new picture and start this again
		popLength = len(df['Cumulative Probs'])
		#Generate boolean array used to keep track of unique samples appended to the set of new samples
		used2 = np.zeros(panoDims[1], dtype=bool)
		uniques = 0
		uniquePoints = []
		
		#Another monitoring tool. Uncomment for data before this point.
		#df.to_csv(dataPath + 'Before.csv', index = False)		
		
		#This loop runs horribly slow because it needs to do so many damned if-checks. Reenable this if you want to see why we decided to go with using Binary Search instead.
		
		print('Population size: ' + str(popLength))
		print('Selecting new population...')
		newPop = []
		for i in range(popLength):
			x = random.random()
			
			#Perform binary search to find the first value that is greater than x faster than iterating through the whole list backwards would be.
			#This stuff is untested, as it's the day of the demo and I only just know had the idea to do this (don't do that, by the way). If this way of finding the desired value is crap, comment it out and uncomment the other loop below to get one that does work, if much slower.
			#...or fix it!
			leftBound = 0
			rightBound = popLength - 1
			index = 0
			while(True):
				index = math.floor((rightBound + leftBound) / 2)
				if(x < df.loc[index]['Cumulative Probs'] and (index == 0 or (x >= df.loc[index - 1]['Cumulative Probs']))):
					break
				elif(x < df.loc[index]['Cumulative Probs']):
					rightBound = index - 1
				else:
					leftBound = index + 1
			
			
			#index = popLength - 1
			#while index > 0 and x < df.loc[index - 1]['Cumulative Probs']:
			#	index -= 1			
			
			member = df.loc[index, 'Population Pixel']
			newPop.append(member)
			#print('Appended point ' + str(df.loc[index]['Population Pixel']))
			if used2[member] == False:
				used2[member] = True
				uniquePoints.append(member)
				uniques += 1
				#print('Unique!/n')
			#print('x selected: ' + str(x))
			
			#You can also get rid of this block here to minimize console printout, if you'd like.
			if((i + 1) % 100 == 0 or (i + 1) == popLength):
				print('Member ' + str(i + 1) + ' added to population')
		
		
		
		df['Population Pixel'] = newPop
		print('Replaced old population with new')
		print('Unique members remaining: ' + str(uniques))
		
		
		#Iterate through the samples taken and count up the unique pixel entries; if the number of uniques is >95% of the total possible samples or you've run through the maximum number of iterations, stop, since you've likely localized yourself; otherwise, run the algorithm again; rinse and repeat until uniques / totalPop > 95%.
		if (uniques / popLength) < .075 or iters >= 10:
			
			#Speen man funni
			print('This is it, Luigi...')
			
			medoid = 0
			for popMember in uniquePoints:
			
				reverse = panoDims[1] - popMember
				farther = (popMember + reverse)
				
				#If farther > 0, it takes /longer to reach the population member by starting at 0 and moving right rather than starting at panoWidth and moving left; vice versa if otherwise
				if farther > 0:
					medoid -= reverse
				else:
					medoid += popMember
			#If medoid = positive, the + panoWidth does nothing; if negative, it wraps around the panorama:
			#Ex: med = -200, panoWidth = 1309
			#-200 + 1309 = 1109; 1109 % 1309 = 1109
			
			#The % panoWidth was intended to act as a safety net in case medoid / uniques was somehow > panoWidth, but I don't think that'll ever happen. Reenable it if you find it does.
			final = (((((medoid / uniques) + panoDims[1]) % panoDims[1]) / panoDims[1]) * 360)
			
			print('Final guess: ' + str(final))
		
		
		#I am unsure if a new .csv file is created every time you save the DataFrame object as one. If it doesn't, you might have junk data left over from your last run, which won't affect the program but would affect the file it spits out at the end.
		#See if there's some way to save the DataFrame as a new file through some function in pandas.
		df.to_csv(dataPath + 'Data.csv', index = False)
		print('Data CSV saved')		
		
		iters += 1
		
	return final

def rotato(robot : cozmo.robot.Robot):
	"""Rotates the robot in ~20 degree intervals as it gathers data to try and localize itself"""
	
	#Enable the robot's camera
	robot.camera.image_stream_enabled = True
	
	#df = varia.getDataFrame()
	
	fileName = varia.getPicturePath() + "latestImage.jpeg"
	rotate = 20
	
	#Rotate in intervals of 20. Assume some random noise in the rotation from mechanical imprecision in the robot and apply that as an update value.
	robot.turn_in_place(cozmo.util.degrees(rotate), cozmo.util.speed_mmps(varia.getTurnSpeed()), varia.getTurnAccel(), ).wait_for_completed()
	
	#Update df poses
	#df['X Updates'] = df['X Updates'].apply(20 + random.uniform(.01, .025))
	
	#Take a picture and save it; downres and blur for faster computing within the localization process
	latest_image = robot.world.latest_image
	annotated = latest_image.annotate_image()
	converted = annotated.convert()
	imgArray = np.asarray(converted)
	scale = varia.getScale()
	downRes = cv2.resize(imgArray, (int(varia.getWidth() * scale), int(varia.getHeight() * scale)))
	blur = cv2.GaussianBlur(downRes, (5, 5), cv2.BORDER_DEFAULT)

	cv2.imwrite(fileName, blur)
	
	#I would normally return this value as I would in any other method, but it seems to get lost if you call cozmo.run_program(something); I guess it's really running two programs and the outermost one doesn't retain this value.
	varia.setUpdate(rotate + random.uniform(.01, .025))
	
#Comment this out once you're done testing the method
#localize()
