#!/usr/bin/env python3

# Copyright (c) 2016 Anki, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the file LICENSE.txt or at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#Implementation of Monte Carlo locolization
from PIL import Image
from sys import platform
import cozmo
import cv2
import numpy as np
from random import randint
import Stitching
import Localize
import MotionUpdate
import Variables as varia

def main():
	cozmo.run_program(readyUp)

	if cv2.imread(varia.getPicturePath() + 'Panorama.jpeg') is None and cv2.imread(varia.getPicturePath() + 'latestImage.jpeg') is None:
		#Calls the initial localizing algorithm to gather data on surroundings
		print('Panorama does not already exist; running full program')
		
		#Initial set up for the panorama, if it doesn't exist
		cozmo.run_program(gatherInfo)
		
		#Run the stitching alrgorithm 
		Stitching.stitch()
	else:
		print('Panorama already exists; skipping information gathering')
	
	#Pause before demo
	input('Program paused; press Enter to continue')
	
	#Turns the robot a random number of degrees, "kidnapping" it
	cozmo.run_program(randomTurn)
	
	#Runs the algorithm until it believes it has found where it's facing in relation to its prior data
	global final
	final = Localize.localize()
	
	#MotionUpdate.motionUpdate()
	cozmo.run_program(finished)

def readyUp(robot: cozmo.robot.Robot):
	"""Sets the robot's desired arm and head orientation to avoid interference"""
	move_arms = robot.set_lift_height(0)
	move_arms.wait_for_completed()

	set_head = robot.set_head_angle((cozmo.robot.MAX_HEAD_ANGLE) / 3, in_parallel = True)
	set_head.wait_for_completed()

def gatherInfo(robot: cozmo.robot.Robot):
	"""Spins the robot 360 degrees to get a panorama image of its current environment"""

	#Roll off of charger
	robot.drive_off_charger_contacts().wait_for_completed()
	
	#robot.say_text("Beginning program.").wait_for_completed()
	print(varia.getPicturePath)


   	#Enable the robot's camera
	robot.camera.image_stream_enabled = True
	
	picturePath = varia.getPicturePath()
	
	#Range = range of rotation until a completed pandorama can be created.
	degree = 0
	turn = varia.getTurn()
	range = varia.getRange()
	while(degree < range) :
	
		#Initalizes the filepath we want each image to be in, different images are seperated by degrees
		#print('Platform? :')
		#print(platform)
		fileName = picturePath + 'stitchpic' + str(degree) + ".jpeg"

		#Robot turns to match degrees and then takes the image
		robot.turn_in_place(cozmo.util.degrees(turn), cozmo.util.speed_mmps(varia.getTurnSpeed()), varia.getTurnAccel()).wait_for_completed()
		
		latest_image = robot.world.latest_image
		
		#Re-enable commented code below for faster image stitching at the cost of accuracy, good for bugtesting
		
		#converted = annotated.convert()
		#imgArray = np.asarray(latest_image.annotate_image().convert())
		#scale = varia.getScale()
		#downRes = cv2.resize(imgArray, (int(varia.getWidth() * scale), int(varia.getHeight() * scale)))
		#sigma = varia.getSigma()
		#blur = cv2.GaussianBlur(downRes, (sigma, sigma), 0)

		cv2.imwrite(fileName, np.asarray(latest_image.annotate_image()))
   
		print("Image %s success" % degree)
		
		degree += turn

		
def randomTurn(robot: cozmo.robot.Robot):
	"""Turns the robot a random number of degrees to simulate the 'Kidnapped Robot Problem'"""

	#Enable the robot's camera
	robot.camera.image_stream_enabled = True
	
	fileName = varia.getPicturePath() + "latestImage.jpeg"

	#Rotate a random number of degrees within a full 360 degree range; setting the ranges in this way is done to minimize any shifting around the robot does with large movements. I figured you could just "sign" the rotational values in this way to do that.
	degree = randint(-180, 180)
	
	#robot.say_text('Random number generated. Rotating' + str(degree) ' degrees from home.').wait_for_completed()
	
	robot.turn_in_place(cozmo.util.degrees(degree), cozmo.util.speed_mmps(varia.getTurnSpeed()), varia.getTurnAccel()).wait_for_completed()
	
	#print('Degrees rotated: ' + str(degree))
	
	#This reset isn't really necessary; it's just here for the sake of saying we don't retain that information.
	degree = 0
	
	#robot.say_text('Memory purged. Where am I?').wait_for_completed()
	
	#Take a picture and save it, note how we modify the image
	latest_image = robot.world.latest_image
	annotated = latest_image.annotate_image()
	converted = annotated.convert()
	imgArray = np.asarray(converted)
	scale = varia.getScale()
	downRes = cv2.resize(imgArray, (int(varia.getWidth() * scale), int(varia.getHeight() * scale)))
	blur = cv2.GaussianBlur(downRes, (5, 5), cv2.BORDER_DEFAULT)

	cv2.imwrite(fileName, blur)


def finished(robot: cozmo.robot.Robot):
	input('Best guess: ' + str(final) + ' degrees from the left side of panorama.')
	#robot.say_text('I''m facing ' + str(final) + ' degrees from the left of my panorama shot. Did I get it right?').wait_for_completed()

main()
