#Code needed to stitch images together 
import cv2
import numpy as np
import sys
import os
from math import floor
from numba import jit
import Variables as varia

#This thing here makes the program use the current computer's GPU to run the code instead. It's supposed to speed it up greatly, when it decides to actually work - sometimes it just seems to take just as long as it would if it were using the CPU. I dunno.
@jit
def stitch():
	"""Initial method used to run program"""
	#Create image array
	images = []
	#print('begin')
	
	picturePath = varia.getPicturePath()
	turn = varia.getTurn()
	#This should calculate the number of stitchable images you have saved.
	stitches = int((varia.getRange() / 10) / (turn / 10))
	
	#As mentioned earlier in Start's gatherInfo, you could potentially downsample each stitchpic for faster stitching here - the catch is, if you do that, you might have issues with stitching them together if they've been downsampled too much, since the stitcher relies on clear detail within the images. Maybe try scaling them down to 95% of their original sizes and testing from there if you really need to get this running faster.
	#Add images to image array
	for i in range(stitches):
		image = cv2.imread(picturePath + 'stitchpic' + str((i * turn)) + ".jpeg")
		images.append(image)
		print('Append ' + str(i * turn) + ' success')
	print('Finished image appending')
	print('Total number of images appended: ' + str(len(images)))
	#print(len(images))
	
	#Stitch images together
	stitcher = cv2.Stitcher.create()
	print("Beginning stitch, hold onto your butts!")
	panoStatus, pano = stitcher.stitch(images)
	print('Stitching completed!')
	
	#Perform a check to make sure the stitching was successful; if so, save the stitched image along with a downres'd version of it.
	if panoStatus == 0:
		cv2.imwrite(picturePath + 'Fullres.jpeg', pano)
		print('Fullres panorama saved!')
		
		#This crop is done to try and get rid of most of those damned black pixels you'll see in the fullres stitched image; this should help with getting more accurate localization results since the MSE comparison relies on the luminosity values of these pictures. Also may help speed up image processing in Localize(), if only (probably) slightly.
		crop = pano[int(pano.shape[0] / 16):pano.shape[0] - int(pano.shape[0] / 16), 0:pano.shape[1]]
		cv2.imwrite(picturePath + 'Cropped.jpeg', crop)
		print('Cropped panorama saved!')
		
		#Downres and blur the image for faster processing later
		scale = varia.getScale()
		downRes = cv2.resize(crop, (int(pano.shape[1] * scale), int(pano.shape[0] * scale)))
		sigma = varia.getSigma()
		blur = cv2.GaussianBlur(downRes, (sigma, sigma), 0)
		cv2.imwrite(picturePath + 'Panorama.jpeg', blur)
		#print('Blurred width: ' + str(blur.shape[1]))
		print('Final blurred panorama saved!')
	else:
		print('Encountered an error creating the panorama\n\n\n\n\n\n\n\n\n\n')
		
#Uncomment this next line to test the method.
#stitch()
